# !/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2024 Commandcracker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from textual.reactive import reactive
from textual.screen import ModalScreen

# built-in modules
from asyncio import gather
from shutil import which
from time import sleep, time
from typing import Union
from subprocess import Popen, PIPE
from os import name as os_name
import sys

# pip modules
# import mpv
# import ytdl
from textual import events, on, work
from random import choice
from textual.app import App, ComposeResult
from textual.containers import Center, Container, Horizontal, ScrollableContainer
from textual.widgets import (Button, Checkbox, DataTable, Footer, Header,
                             Input, Label, ListItem, ListView, Markdown,
                             TabbedContent, TabPane, Static, Switch, RadioButton)
from textual.widgets._toggle_button import ToggleButton

from .provider.common import SearchResult, Series, Episode, Language
from .provider.aniworld import AniWorldProvider
from .provider.serienstream import SerienStreamProvider
from .player.common import Player
from .player.mpv import MPVPlayer
from .player.vlc import VLCPlayer
from .player.wmplayer import WMPlayer
from .player.ffplay import FFPlayPlayer
from .player.android import AndroidVLCPlayer, AndroidMPVPlayer, AndroidChoosePlayer
from .hoster.common import Hoster, DirectLink
from .hoster.veo import VOEHoster
from .hoster.vidoza import VidozaHoster
from .hoster.streamtape import StreamtapeHoster
from .hoster.doodstream import DoodstreamHoster
from .aniskip import (
    get_timings_from_search,
    timings_to_mpv_options,
    generate_chapters_file_and_get_mpv_option
)

import logging
from pathlib import Path

logs_path = Path(__file__).parent.parent.parent.joinpath("logs")
logs_path.mkdir(exist_ok=True, parents=True)
logging.basicConfig(filename=logs_path.joinpath("gucken.log"), encoding='utf-8', level=logging.INFO)


class Next(ModalScreen):
    time = reactive(3)

    def __init__(self, question: str, no_time: bool = False):
        super().__init__()
        self.question = question
        self.no_time = no_time

    def compose(self) -> ComposeResult:
        with Container():
            yield Label(self.question)
            with Horizontal():
                yield Button.error("Cancel", id="cancel")
                yield Button.success("Next", id="next")

    def on_mount(self) -> None:
        if not self.no_time:
            self.set_interval(1, self.update_time)
            self.query_one(Label).update(self.question + " " + str(self.time))

    def update_time(self) -> None:
        self.time = self.time - 1
        if self.time < 0:
            self.dismiss(True)
        self.query_one(Label).update(self.question + " " + str(self.time))

    @on(Button.Pressed)
    def exit_screen(self, event):
        button_id = event.button.id
        self.dismiss(button_id == "next")


class ClickableListItem(ListItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_click = None

    def on_click(self) -> None:
        if self.last_click and time() - self.last_click < 0.5:
            self.app.open_info()
        self.last_click = time()


class ClickableDataTable(DataTable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_click = {}

    def on_click(self, event: events.Click) -> None:
        meta = event.style.meta
        if not "row" in meta or not "column" in meta:
            return
        row_index = meta["row"]
        if row_index <= -1:
            return
        if self.last_click.get(row_index) and time() - self.last_click[row_index] < 0.5:
            self.app.play_selected()
        self.last_click[row_index] = time()


class GuckenApp(App):
    TITLE = "Gucken TUI"
    # TODO: color theme https://textual.textualize.io/guide/design/#designing-with-colors
    CSS_PATH = "gucken.tcss"

    def __init__(self):
        super().__init__()
        self.current: Union[list[SearchResult], None] = None
        self.current_info: Union[Series, None] = None

    def compose(self) -> ComposeResult:
        yield Header()
        with TabbedContent():
            with TabPane("Search", id="search"):  # Search "🔎"
                with Horizontal(id="hosters"):
                    yield Checkbox("AniWorld.to", True, id="aniworld_to")
                    yield Checkbox("SerienStream.to", id="serienstream_to")
                yield Input(id="input", placeholder="Search for a Anime")
                yield ListView(id="results")
            with TabPane("Info", id="info", disabled=True):  # Info "ℹ"
                with ScrollableContainer(id="res_con"):
                    yield Markdown(id="markdown")
                    yield ClickableDataTable(id="season_list")
            with TabPane("Settings", id="setting"):  # Settings "⚙"
                yield RadioButton("Fullscreen", id="fullscreen", value=True)
                yield RadioButton("Syncplay", id="syncplay", value=False)
                yield RadioButton("ani-skip", id="ani_skip", value=True)
            # with RadioSet():
            #    yield RadioButton("VOE", id="voe", value=True)
            #    yield RadioButton("Doodstream", id="doodstream")
            #    yield RadioButton("Vidoza", id="vidoza")
            #    yield RadioButton("Streamtape", id="streamtape")
        with Footer():
            with Center():
                yield Label("Made by Commandcracker with ❤")

    def on_mount(self) -> None:
        self.query_one(Input).focus()
        self.query_one("#info", TabPane).set_loading(True)
        table = self.query_one(DataTable)
        table.cursor_type = "row"
        # TODO: make them scale
        table.add_columns("FT", "S", "F", "Name", "Title", "Hoster", "Sprache")
        if not which("mpv"):
            self.notify(
                "You wont be able to play videos directly.\n"
                "Please install MPV!",
                title="MPV not found",
                severity="warning",
            )

    async def on_checkbox_changed(self):
        self.lookup_anime(self.query_one("#input", Input).value)

    async def on_input_changed(self, message: Input.Changed) -> None:
        if message.value:
            self.lookup_anime(message.value)
        else:
            await self.query_one("#results", ListView).clear()

    # TODO: https://textual.textualize.io/guide/workers/#thread-workers
    @work(exclusive=True)
    async def lookup_anime(self, keyword: str) -> None:

        owos = []
        if self.query_one("#aniworld_to", Checkbox).value:
            owos.append(AniWorldProvider.search(keyword))

        if self.query_one("#serienstream_to", Checkbox).value:
            owos.append(SerienStreamProvider.search(keyword))

        lv = self.query_one("#results", ListView)
        await lv.clear()
        await lv.set_loading(True)
        results = await gather(*owos)
        f_results = []
        for l in results:
            if l is not None:
                for e in l:
                    f_results.append(e)

        # TODO: stor f_results by fuzzy sort keyword
        if len(f_results) > 0:
            self.current = f_results
            for series in f_results:
                await lv.append(ClickableListItem(Markdown(
                    f"##### **{series.name}** {series.production_year}\n{series.description}"
                )))
        await lv.set_loading(False)
        if len(f_results) > 0:
            lv.index = 0

    async def on_key(self, event: events.Key) -> None:
        key = event.key
        if self.screen.id == "_default":
            if self.query_one(TabbedContent).active == "search":
                lv = self.query_one("#results", ListView)
                inp = self.query_one("#input", Input)
                # Down to list
                if key == "down":
                    if inp.has_focus:
                        lv.focus()
                # Up to Input
                if key == "up":
                    if not inp.has_focus and (lv.index == 0 or lv.index is None):
                        inp.focus()
                # Selection
                if key == "enter":
                    if lv.index is not None:
                        self.open_info()
                # Type anywhere
                if key not in ["down", "up", "enter"]:
                    if lv.has_focus:
                        inp.focus()
                        if key == "backspace":
                            inp.action_delete_left()
                        else:
                            await inp.on_event(event)
            if key == "enter" and self.query_one(DataTable).has_focus:
                self.play_selected()

    def sort_favorite_lang(self, language_list: list[Language]) -> list[Language]:
        return sorted(language_list, key=lambda x: (
            x != Language.DE,
            x != Language.JP_DESUB,
            x != Language.JP_ENSUB,
        ))

    def sort_favorite_hoster(self, hoster_list: list[Hoster]) -> list[Hoster]:
        return sorted(hoster_list, key=lambda x: (
            not isinstance(x, StreamtapeHoster),
            not isinstance(x, VOEHoster),
            not isinstance(x, VidozaHoster),
            not isinstance(x, DoodstreamHoster),
        ))

    async def get_working_direct_link(self, hosters: list[Hoster]) -> Union[DirectLink, None]:
        for hoster in hosters:
            direct_link = await hoster.get_direct_link()
            is_working = await direct_link.check_is_working()
            logging.info('Check: "%s" Working: "%s" URL: "%s"', type(hoster).__name__, is_working, direct_link)
            if is_working:
                return direct_link
        return None

    @work(exclusive=True)
    async def play_selected(self):
        # TODO: cache more
        # TODO: dont only check for mpv
        #if which("mpv"):
        dt = self.query_one(DataTable)
        # TODO: show loading
        await dt.set_loading(True)
        index = self.app.query_one("#results", ListView).index
        series_search_result = self.current[index]
        self.play(
                series_search_result=series_search_result,
                episodes=self.current_info.episodes,
                index=dt.cursor_row
            )
        await dt.set_loading(False)
        #else:
        #    self.notify(
        #        "You wont be able to play videos directly.\n"
        #        "Please install MPV!",
        #        title="MPV not found",
        #        severity="error",
        #    )

    @work(exclusive=True)
    async def open_info(self) -> None:
        series_search_result: SearchResult = self.current[self.app.query_one("#results", ListView).index]
        info_tab = self.query_one("#info", TabPane)
        info_tab.disabled = False
        self.query_one(TabbedContent).active = "info"
        md = self.query_one("#markdown", Markdown)
        await info_tab.set_loading(True)
        series = await series_search_result.get_series()
        self.current_info = series
        await md.update(series.to_markdown())

        table = self.query_one(DataTable)
        table.clear()
        c = 0
        for ep in series.episodes:
            hl = []
            for h in ep.available_hoster:
                if h is VOEHoster:
                    hl.append("VEO")
                if h is DoodstreamHoster:
                    hl.append("D")
                if h is VidozaHoster:
                    hl.append("VZ")
                if h is StreamtapeHoster:
                    hl.append("ST")

            ll = []
            for l in ep.available_language:
                if l == Language.DE:
                    ll.append("DE")
                if l == Language.JP_DESUB:
                    ll.append("JP_DESUB")
                if l == Language.JP_ENSUB:
                    ll.append("JP_ENSUB")

            c += 1
            table.add_row(
                c,
                ep.season,
                ep.episode_number,
                ep.title,
                "",
                ", ".join(hl),
                ", ".join(ll)
            )
        table.focus(scroll_visible=False)
        await info_tab.set_loading(False)

    def detect_player(self) -> Union[Player, None]:
        # Android
        if hasattr(sys, 'getandroidapilevel'):
            # TODO: detect right
            return AndroidMPVPlayer()
            #return AndroidVLCPlayer()
            #return AndroidChoosePlayer()
            #return None

        # All
        if which("mpv"):
            return MPVPlayer()
        if which("vlc"):
            return VLCPlayer()

        # Windows
        if os_name == "nt":
            if which(r"C:\Program Files\VideoLAN\VLC\vlc.exe"):
                # r"C:\Program Files\VideoLAN\VLC\vlc.exe"
                return VLCPlayer()

        if which("ffplay"):
            return FFPlayPlayer()

        # Windows
        if os_name == "nt":
            return WMPlayer()

        return None

    @work(thread=True)
    async def play(
            self,
            series_search_result: SearchResult,
            episodes: list[Episode],
            index: int
    ) -> None:
        episode: Episode = episodes[index]
        processed_hoster = await episode.process_hoster()

        lang = self.sort_favorite_lang(episode.available_language)[0]
        sorted_hoster = self.sort_favorite_hoster(processed_hoster.get(lang))
        direct_link = await self.get_working_direct_link(sorted_hoster)

        player = self.detect_player()

        if player is None:
            raise RuntimeError()

        # TODO: ani_skip, syncplay, fullscreen as arg
        # TODO: pass ani_skip as script
        ani_skip = self.query_one("#ani_skip", RadioButton).value
        syncplay = self.query_one("#syncplay", RadioButton).value
        fullscreen = self.query_one("#fullscreen", RadioButton).value

        title = f"{series_search_result.name} - {episode.title}"

        path = None
        if isinstance(player, VLCPlayer):
            if not which("vlc"):
                if which(r"C:\Program Files\VideoLAN\VLC\vlc.exe"):
                    path = r"C:\Program Files\VideoLAN\VLC\vlc.exe"

        args = player.play(direct_link.url, title, fullscreen, direct_link.headers, path)

        if isinstance(player, MPVPlayer):
            if ani_skip:
                timings = await get_timings_from_search(series_search_result.name, index + 1)
                if timings:
                    # --start=00:56
                    args += [
                        timings_to_mpv_options(timings),
                        generate_chapters_file_and_get_mpv_option(timings)
                    ]

        if syncplay:
            syncplay_path = None
            if which("syncplay"):
                syncplay_path = "syncplay"
            if not syncplay_path:
                if os_name == "nt":
                    if which(r"C:\Program Files (x86)\Syncplay\Syncplay.exe"):
                        syncplay_path = r"C:\Program Files (x86)\Syncplay\Syncplay.exe"
            if not syncplay_path:
                self.notify(
                    "Syncplay not found",
                    title="Syncplay not found",
                    severity="error",
                )
            else:
                if isinstance(player, MPVPlayer) or isinstance(player, VLCPlayer):
                    # TODO: dont detect mpv.com
                    player_path = which(args[0])
                    url = args[1]
                    args.pop(0)
                    args.pop(0)
                    args = [syncplay_path, "--player-path", player_path, url, "--"] + args
                else:
                    self.notify(
                        "Your player is not supported by Syncplay",
                        title="Player not supported",
                        severity="warning",
                    )

        logging.info("Running: %s", args)
        process = Popen(
            args,
            stderr=PIPE
        )
        while not self.app._exit:
            sleep(0.1)

            resume_time = None

            # only if mpv
            while not self.app._exit:
                output = process.stderr.readline()
                if process.poll() is not None:
                    break
                if output:
                    out_s = output.strip().decode()
                    # AV: 00:11:57 / 00:24:38 (49%) A-V:  0.000 Cache: 89s/22MB
                    if out_s.startswith("AV:"):
                        sp = out_s.split(" ")
                        resume_time = sp[1]

            if resume_time:
                logging.info("Resume: %s", resume_time)

            exit_code = process.poll()

            if exit_code is not None:

                async def push_next_screen():
                    async def play_next(should_next):
                        if should_next:
                            self.play(
                                series_search_result,
                                episodes,
                                index + 1,
                            )

                    await self.app.push_screen(Next("Playing next episode in", no_time=hasattr(sys, 'getandroidapilevel')), callback=play_next)

                self.app.call_later(push_next_screen)
                return


exit_quotes = [
    "Closing one anime is just an invitation to open another.",
    "You finished one, now finish the next.",
    "Don't stop now, there's a whole universe waiting to be explored.",
    "The end of one journey is just the beginning of another.",
    "Like a phoenix rising from the ashes, the end of one episode ignites the flames of anticipation for the next."
]


def main():
    gucken_app = GuckenApp()
    gucken_app.run()
    print(choice(exit_quotes))


if __name__ == "__main__":
    main()



from httpx import AsyncClient


async def search(keyword: str) -> dict:
    async with AsyncClient(verify=False) as client:
        response = await client.get(f"https://myanimelist.net/search/prefix.json?type=anime&keyword={keyword}")
        return response.json()



from re import compile as re_compile
from httpx import AsyncClient
from .common import Hoster, DirectLink

# TODO: improve all patterns
EXTRACT_VIDOZA_HLS_PATTERN = re_compile(r"sourcesCode:.*?\[.*?\{.*?src:.*?[\'|\"](?P<hls>.*?)[\'|\"],")


class VidozaHoster(Hoster):
    async def get_direct_link(self) -> DirectLink:
        async with AsyncClient(verify=False) as client:
            response = await client.get(self.url, follow_redirects=True)
            match_hls = EXTRACT_VIDOZA_HLS_PATTERN.search(response.text)
            return DirectLink(match_hls.group(1))



from abc import abstractmethod, ABC
from dataclasses import dataclass
from enum import Enum
from typing import Union
from ..hoster.common import Hoster


class Language(Enum):
    DE = "Deutsch"
    JP_DESUB = "Japanisch mit deutschen untertitel"
    JP_ENSUB = "Japanisch mit englischen untertitel"


@dataclass
class Episode:
    title: str
    season: int
    episode_number: int
    total_episode_number: int
    # language: list[Language]
    # hoster: set[Hoster]
    # title_en: str
    # title_de: Union[str, None]
    # language: set[Language]
    # hoster: set[Hoster]
    # number: int

    available_language: list[Language]
    available_hoster: list[Hoster]

    @abstractmethod
    async def process_hoster(self) -> dict[Language, list[Hoster]]:
        raise NotImplementedError


@dataclass
class Series:
    name: str
    episodes: list[Episode]

    # description: str

    @abstractmethod
    def to_markdown(self) -> str:
        raise NotImplementedError


@dataclass
class SearchResult:
    name: str
    description: str = None

    @abstractmethod
    async def get_series(self) -> Series:
        raise NotImplementedError

    @property
    @abstractmethod
    def url(self) -> str:
        raise NotImplementedError


class Provider(ABC):

    @staticmethod
    @property
    @abstractmethod
    def host() -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    async def search(keyword: str) -> Union[list[SearchResult]]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    async def get_series(search_result: SearchResult) -> Series:
        raise NotImplementedError



from .common import Player


class WMPlayer(Player):
    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        args = [
            override_executable or r"C:\Program Files (x86)\Windows Media Player\wmplayer.exe",
            url
        ]
        if full_screen:
            args.append("/fullscreen")
        return args



from .common import Player, dict_to_string


class FFPlayPlayer(Player):
    supports_headers = True

    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        args = [
            override_executable or "ffplay",
            url
        ]
        if title:
            args.append("-window_title")
            args.append(title)
        if headers:
            args.append(f"-headers={dict_to_string(headers)}")
        return args



from .common import Player


class AndroidChoosePlayer(Player):
    supports_headers = False

    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        return [
            "am",
            "start",
            "--user",
            "0",
            "-a"
            "android.intent.action.VIEW",
            "-d",
            url,
            "-c",
            "android.intent.category.BROWSABLE"
        ]


# http://mpv-android.github.io/mpv-android/intent.html
class AndroidMPVPlayer(Player):
    supports_headers = False

    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        return [
            "am",
            "start",
            "--user",
            "0",
            "-a"
            "android.intent.action.VIEW",
            "-d",
            url,
            "-n",
            "is.xyz.mpv/.MPVActivity"
        ]


# https://wiki.videolan.org/Android_Player_Intents/
class AndroidVLCPlayer(Player):
    supports_headers = False

    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        args = [
            "am",
            "start",
            "--user",
            "0",
            "-a"
            "android.intent.action.VIEW",
            "-d",
            url,
            "-n",
            "org.videolan.vlc/org.videolan.vlc.gui.video.VideoPlayerActivity"
        ]
        if title:
            args.append("-e")
            args.append("title")
            args.append(title)
        return args



from httpx import AsyncClient


SEARCH_QUERY = """
query ($id: Int, $page: Int, $perPage: Int, $search: String) {
  Page (page: $page, perPage: $perPage) {
    pageInfo {
      total
      currentPage
      lastPage
      hasNextPage
      perPage
    }
    media (id: $id, search: $search) {
      id
      title {
        romaji
        english
        native
      }
    }
  }
}
"""


async def search(keyword: str) -> dict:
    async with AsyncClient(verify=False) as client:
        response = await client.post(
            f"https://graphql.anilist.co",
            headers={"Content-Type": "application/json"},
            json={"query": SEARCH_QUERY, "variables": {"search": keyword}}
        )
        return response.json()



from .common import Player, dict_to_string


class MPVPlayer(Player):
    supports_headers = True

    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        args = [
            override_executable or "mpv",
            url
        ]
        if full_screen:
            args.append("--fullscreen")
        if title:
            args.append(f"--force-media-title={title}")
        if headers:
            args.append(f"--http-header-fields={dict_to_string(headers)}")
        return args



from re import compile as re_compile
from httpx import AsyncClient
from .common import Hoster, DirectLink

STREAMTAPE_PATTERN = re_compile(r'botlink(.*?)innerHTML(.*?)\);')
STREAMTAPE_PATTERN_SUBSTRING = re_compile(r'substring\(\d+')
STREAMTAPE_PATTERN_DIGETS = re_compile(r'\d+')


class StreamtapeHoster(Hoster):
    async def get_direct_link(self) -> DirectLink:
        # TODO: Error checking
        async with AsyncClient(verify=False) as client:
            response = await client.get(self.url, follow_redirects=True)
            # TODO: Save html and error in order to investigate
            # with open("out.txt", "wb") as f:
            #    f.write(response.text.encode('utf-8'))
            video_src = STREAMTAPE_PATTERN.search(response.text)
            j1 = "".join(video_src.groups())
            u1 = j1.split(" ")[2][1:-2]
            u2 = j1[j1.index("('") + 2:j1.rfind("')")]

            matches = STREAMTAPE_PATTERN_SUBSTRING.findall(j1)
            for match in matches:
                sub = STREAMTAPE_PATTERN_DIGETS.search(match).group(0)
                u2 = u2[int(sub):]

            return DirectLink(f"https:{u1}{u2}")



from abc import abstractmethod
from dataclasses import dataclass


def dict_to_string(dictionary):
    return ','.join([f"{key}: {value}" for key, value in dictionary.items()])


@dataclass
class Player:
    supports_headers: bool = False

    @abstractmethod
    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        raise NotImplementedError



from httpx import AsyncClient
from .tracker.myanimelist import search
from difflib import SequenceMatcher
from typing import Union
from tempfile import TemporaryFile


# TODO: improve fuzzy

def fuzzy_search(pattern, possibilities, threshold=0.6):
    matches = []
    for word in possibilities:
        ratio = SequenceMatcher(None, pattern, word).ratio()
        if ratio >= threshold:
            matches.append((word, ratio))
    return matches


def fuzzy_sort(pattern, possibilities):
    return sorted(possibilities, key=lambda x: SequenceMatcher(None, pattern, x).ratio(), reverse=True)


async def get_timings_from_id(anime_id: int, episode_number: int) -> Union[dict[str, float], None]:
    async with AsyncClient(verify=False) as client:
        response = await client.get(
            f"https://api.aniskip.com/v1/skip-times/{anime_id}/{episode_number}?types=op&types=ed"
        )
        json = response.json()
        if json.get("found") is not True:
            return None
        op_start_time = 0
        op_end_time = 0
        ed_start_time = 0
        ed_end_time = 0
        for result in json["results"]:
            skip_type = result["skip_type"]
            start_time = result["interval"]["start_time"]
            end_time = result["interval"]["end_time"]
            if skip_type == "op":
                op_start_time = start_time
                op_end_time = end_time
            if skip_type == "ed":
                ed_start_time = start_time
                ed_end_time = end_time
        return {
            "op_start_time": float(op_start_time),
            "op_end_time": float(op_end_time),
            "ed_start_time": float(ed_start_time),
            "ed_end_time": float(ed_end_time)
        }


async def get_timings_from_search(keyword: str, episode_number: int) -> Union[dict[str, float], None]:
    # TODO: improve search
    myanimelist_search_result = await search(keyword)
    animes = {}
    for anime in myanimelist_search_result["categories"][0]["items"]:
        animes[anime["name"]] = anime["id"]
    search_results = fuzzy_search(keyword, animes)
    if len(search_results) > 0:
        name = search_results[0][0]
        anime_id = animes[name]
        return await get_timings_from_id(anime_id, episode_number)
    return None


def timings_to_mpv_options(timings=dict[str, float]) -> str:
    op_start_time = timings["op_start_time"]
    op_end_time = timings["op_end_time"]
    ed_start_time = timings["ed_start_time"]
    ed_end_time = timings["ed_end_time"]
    return f"--script-opts=skip-op_start={op_start_time},skip-op_end={op_end_time},skip-ed_start={ed_start_time},skip-ed_end={ed_end_time}"


def chapter(start: float, end: float, title: str) -> str:
    return f"\n[CHAPTER]\nTIMEBASE=1/1000\nSTART={int(start * 1000)}\nEND={int(end * 1000)}\nTITLE={title}\n"


def get_chapters_file_content(timings=dict[str, float]) -> str:
    op_start_time = timings["op_start_time"]
    op_end_time = timings["op_end_time"]
    ed_start_time = timings["ed_start_time"]
    ed_end_time = timings["ed_end_time"]
    return (
            ";FFMETADATA1" +
            chapter(op_start_time, op_end_time, "Opening") +
            chapter(ed_start_time, ed_end_time, "Ending") +
            chapter(op_end_time, ed_start_time, "Episode")
    )


def generate_chapters_file(timings=dict[str, float]) -> TemporaryFile:
    temp_file = TemporaryFile(mode='w', prefix="gucken-", delete=False)
    temp_file.write(get_chapters_file_content(timings))
    return temp_file


def get_chapters_file_mpv_option(path: str) -> str:
    return f"--chapters-file={path}"


def generate_chapters_file_and_get_mpv_option(timings=dict[str, float]) -> str:
    return get_chapters_file_mpv_option(generate_chapters_file(timings).name)



from base64 import b64decode
from re import compile as re_compile
from httpx import AsyncClient
from .common import Hoster, DirectLink

EXTRACT_VEO_HLS_PATTERN = re_compile(r"'hls': '(.*?)'")


class VOEHoster(Hoster):
    async def get_direct_link(self) -> DirectLink:
        async with AsyncClient(verify=False) as client:
            response = await client.get(self.url, follow_redirects=True)
            match_hls = EXTRACT_VEO_HLS_PATTERN.search(response.text)
            hls_link = match_hls.group(1)
            return DirectLink(b64decode(hls_link).decode())



from .common import Player


class VLCPlayer(Player):
    def play(
            self,
            url: str,
            title: str,
            full_screen: bool,
            headers: dict[str, str] = None,
            override_executable: str = None
    ) -> list[str]:
        args = [
            override_executable or "vlc",
            url,
            "--no-video-title-show",
            "--play-and-exit",
        ]
        if full_screen:
            args.append("--fullscreen")
        if title:
            args.append(f"--input-title-format={title}")
        if headers:
            if headers.get("Referer"):
                args.append(f"--http-referrer={headers.get('Referer')}")
            if headers.get("User-Agent"):
                args.append(f"--http-user-agent={headers.get('User-Agent')}")
        return args



from abc import abstractmethod
from dataclasses import dataclass, field
from httpx import HTTPError, AsyncClient


@dataclass
class DirectLink:
    url: str
    headers: dict[str, str] = None

    async def check_is_working(self) -> bool:
        try:
            async with AsyncClient(verify=False) as client:
                response = await client.head(self.url, follow_redirects=True, headers=self.headers)
                return response.is_success
        except HTTPError:
            return False

    def __str__(self):
        return self.url

    def has_headers(self) -> bool:
        if self.headers is not None:
            return True
        return False


@dataclass
class Hoster:
    url: str
    requires_headers: bool = field(default=False, init=False)

    @abstractmethod
    async def get_direct_link(self) -> DirectLink:
        raise NotImplementedError



from asyncio import gather
from dataclasses import dataclass
from typing import Union
from html import unescape

from httpx import AsyncClient
from bs4 import BeautifulSoup

from .common import (
    Provider,
    SearchResult,
    Series,
    Episode,
    Hoster,
    Language
)
from ..hoster.doodstream import DoodstreamHoster
from ..hoster.streamtape import StreamtapeHoster
from ..hoster.veo import VOEHoster
from ..hoster.vidoza import VidozaHoster


def data_lang_key_to_lang(data_lang_key: str) -> Language:
    if data_lang_key == "1":
        return Language.DE
    if data_lang_key == "2":
        return Language.JP_ENSUB
    if data_lang_key == "3":
        return Language.JP_DESUB


def provider_to_hoster(provider: str, url: str) -> Hoster:
    if provider == "VOE":
        return VOEHoster(url)
    if provider == "Doodstream":
        return DoodstreamHoster(url)
    if provider == "Vidoza":
        return VidozaHoster(url)
    if provider == "Streamtape":
        return StreamtapeHoster(url)


@dataclass
class SerienStreamEpisode(Episode):
    url: str

    async def process_hoster(self) -> dict[Language, list[Hoster]]:
        async with AsyncClient(verify=False) as client:
            response = await client.get(f"{self.url}/staffel-{self.season}/episode-{self.episode_number}")
            soup = BeautifulSoup(response.text, "html.parser")
            watch_episode = soup.find_all(
                "li",
                class_=lambda value: value and value.startswith("episodeLink"),
                attrs={
                    "data-lang-key": True,
                    "data-link-id": True,
                    "data-link-target": True,
                    "data-external-embed": True
                }
            )

            processed_hoster = {
                Language.DE: list(),
                Language.JP_DESUB: list(),
                Language.JP_ENSUB: list(),
            }

            for episode in watch_episode:
                data_link_id = episode.attrs.get("data-link-id")
                provider = episode.find_next("h4").text

                data_lang_key = episode.attrs.get("data-lang-key")
                lang = data_lang_key_to_lang(data_lang_key)

                hoster = provider_to_hoster(provider, f"https://{SerienStreamProvider.host}/redirect/{data_link_id}")

                processed_hoster[lang].append(hoster)

            return processed_hoster


@dataclass
class SerienStreamSeries(Series):
    regisseure: set[str]
    schauspieler: set[str]
    produzent: set[str]
    land: list[str]
    full_description: str
    production_year: str
    tags: set[str]

    def to_markdown(self) -> str:
        return (
            f"# {self.name} {self.production_year}\n{self.full_description}\n\n"
            f"**Regisseure**: {', '.join(self.regisseure)}\\\n"
            f"**Schauspieler**: {', '.join(self.schauspieler)}\\\n"
            f"**Produzent**: {', '.join(self.produzent)}\\\n"
            f"**Land**: {', '.join(self.land)}\n\n"
            f"**Tags**: {', '.join(self.tags)}"
        )


@dataclass
class SerienStreamSearchResult(SearchResult):
    link: str = None
    cover: str = None
    production_year: str = None
    host: str = None

    async def get_series(self) -> SerienStreamSeries:
        return await SerienStreamProvider.get_series(self)

    @property
    def url(self) -> str:
        return f"https://{self.host}/serie/stream/{self.link}"


class SerienStreamProvider(Provider):
    host = "186.2.175.5"

    @staticmethod
    async def search(keyword: str) -> Union[list[SerienStreamSearchResult], None]:
        if keyword.strip() == "":
            return None
        async with AsyncClient(verify=False) as client:
            response = await client.get(f"https://{SerienStreamProvider.host}/ajax/seriesSearch?keyword={keyword}")
            results = response.json()
            search_results = []
            for series in results:
                search_results.append(SerienStreamSearchResult(
                    name=unescape(series.get("name")),
                    link=series.get("link"),
                    description=unescape(series.get("description")),
                    cover=series.get("cover"),
                    production_year=series.get("productionYear"),
                    host=SerienStreamProvider.host
                ))
            return search_results

    @staticmethod
    async def get_series(search_result: SerienStreamSearchResult) -> SerienStreamSeries:
        async with (AsyncClient(verify=False) as client):
            response = await client.get(search_result.url)
            soup = BeautifulSoup(response.text, "html.parser")

            tags = []
            for genre in soup.find_all("a", class_="genreButton clearbutton", attrs={"itemprop": "genre"}):
                tags.append(genre.text)

            actors = []
            for actor in soup.find_all("li", attrs={"itemprop": "actor"}):
                actors.append(actor.find_next("span", attrs={"itemprop": "name"}).text)

            creators = []
            for creator in soup.find_all("li", attrs={"itemprop": "creator"}):
                creators.append(creator.find_next("span", attrs={"itemprop": "name"}).text)

            countrys = []
            for country in soup.find_all("li", attrs={"data-content-type": "country"}):
                countrys.append(country.find_next("span", attrs={"itemprop": "name"}).text)

            directors = []
            for director in soup.find_all("li", attrs={"itemprop": "director"}):
                directors.append(director.find_next("span", attrs={"itemprop": "name"}).text)

            funcs = []

            staffeln = soup.find("div", attrs={"id": "stream"}, class_="hosterSiteDirectNav").find("ul").find_all("a")
            count = 0
            for staffel in staffeln:
                # TODO: Filme
                if staffel.text != "Filme":
                    count += 1
                    if count > 1:
                        funcs.append(
                            get_episodes_from_url(count, search_result.url))

            eps = await gather(
                get_episodes_from_soup(1, search_result.url, soup),
                *funcs
            )

            feps = []
            for e in eps:
                for b in e:
                    feps.append(b)

            return SerienStreamSeries(
                # cover=f"https://{search_result.host}" + soup.find("div", class_="seriesCoverBox").find("img").attrs.get("data-src"),
                name=unescape(soup.find("h1", attrs={"itemprop": "name"}).find("span").text),
                production_year=unescape(soup.find("div", class_="series-title").find("small").text).lstrip(),
                # age=int(soup.find("div", class_="fsk").find("span").text),
                # imdb_link=soup.find("a", class_="imdb-link").attrs.get("href"),
                full_description=unescape(soup.find("p", class_="seri_des").attrs.get("data-full-description")),
                regisseure=directors,
                schauspieler=actors,
                produzent=creators,
                land=countrys,
                tags=tags,
                # rating_value=int(soup.find("span", attrs={"itemprop": "ratingValue"}).text),
                # rating_count=int(soup.find("span", attrs={"itemprop": "ratingCount"}).text),
                # staffeln=count,
                episodes=feps
            )


async def get_episodes_from_url(staffel: int, url: str) -> list[Episode]:
    async with AsyncClient(verify=False) as client:
        response = await client.get(f"{url}/staffel-{staffel}")
        return await get_episodes_from_page(staffel, url, response.text)


async def get_episodes_from_page(staffel: int, url: str, page: str) -> list[Episode]:
    return await get_episodes_from_soup(staffel, url, BeautifulSoup(page, "html.parser"))


async def get_episodes_from_soup(staffel: int, url: str, soup: BeautifulSoup) -> list[Episode]:
    episodes = []
    e_count = 0
    for episode in soup.find("table", class_="seasonEpisodesList").find("tbody").find_all("tr"):
        title = episode.find_next("td", class_="seasonEpisodeTitle")

        language = set()
        for flag in episode.find_next("td", class_="editFunctions").find_all("img"):
            t = flag.attrs.get("title")
            if t == "Englisch":
                language.add(Language.JP_ENSUB)
            if t == "Deutsch/German":
                language.add(Language.DE)
            if t == "Mit deutschem Untertitel":
                language.add(Language.JP_DESUB)

        hoster = set()
        for h in episode.find_all_next("i", class_="icon"):
            t = h.attrs.get("title")
            if t == "VOE":
                hoster.add(VOEHoster)
            if t == "Doodstream":
                hoster.add(DoodstreamHoster)
            if t == "Vidoza":
                hoster.add(VidozaHoster)
            if t == "Streamtape":
                hoster.add(StreamtapeHoster)

        e_count += 1
        title_en = title.find("span").text.lstrip()
        title_de = title.find("strong").text.lstrip()

        episodes.append(SerienStreamEpisode(
            url=url,
            title=f"{title_en} - {title_de}",
            season=staffel,
            episode_number=e_count,
            available_hoster=hoster,
            available_language=language,
            total_episode_number=None
            #language=language,
            #hoster=hoster,
            #number=e_count,
            #staffel=staffel
        ))

    return episodes



from dataclasses import dataclass

from httpx import AsyncClient
from time import time
from random import choices
from string import ascii_letters, digits
from re import compile as re_compile
from urllib.parse import urlparse
from .common import Hoster, DirectLink

EXTRACT_DOODSTREAM_HLS_PATTERN = re_compile(r"/pass_md5/[\w-]+/[\w-]+")


def random_str(length: int = 10) -> str:
    return ''.join(choices(ascii_letters + digits, k=length))


def js_date_now() -> int:
    return int(time() * 1000)


headers = {'Referer': 'https://d0000d.com/'}


class DoodstreamHoster(Hoster):
    requires_headers = True

    async def get_direct_link(self) -> DirectLink:
        async with (AsyncClient(verify=False) as client):
            response = await client.head(self.url)
            if response.has_redirect_location:
                u2 = urlparse(response.headers.get("Location"))._replace(netloc="d000d.com").geturl()
                response = await client.get(u2)

            pass_md5 = EXTRACT_DOODSTREAM_HLS_PATTERN.search(response.text)
            response = await client.get(f"https://d0000d.com{pass_md5.group()}",
                                        headers={'Referer': 'https://d0000d.com/'})
            return DirectLink(
                url=f"{response.text}{random_str()}?token={pass_md5.group().split('/')[-1]}&expiry={js_date_now()}",
                headers=headers
            )
